﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

using OpenCvSharp;

namespace TrainingImageCapture
{
    public partial class MainForm : Form
    {
        private bool _captureFlg;

        private Bitmap _bmpDisp = null;
        private Graphics _grpDisp = null;

        private Bitmap _bmpCamera = null;
        private bool _saveFlg;

        private Rectangle _rect;

        private uint _imageIndex;

        public MainForm()
        {
            InitializeComponent();
        }

        ///////////////////////////////////////////////////////////////
        //
        // フォームイベント
        //
        ///////////////////////////////////////////////////////////////

        private void MainForm_Load(object sender, EventArgs e)
        {
            _captureFlg = true;
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            _captureFlg = false;

            Properties.Settings.Default.Save();
        }

        private void MainForm_Shown(object sender, EventArgs e)
        {
            var capture = new VideoCapture(0);

            using (var mat = new Mat())
            {
                capture.Read(mat);

                picImage.Width = mat.Cols;
                picImage.Height = mat.Rows;

                // 撮影領域の設定（左側パネルの中心）
                SetROI(splitContainer.Panel1.Width / 2, splitContainer.Panel1.Height / 2);

                // Graphicsの確保
                _bmpDisp = new Bitmap(mat.Cols, mat.Rows, System.Drawing.Imaging.PixelFormat.Format24bppRgb);
                picImage.Image = _bmpDisp;
                _grpDisp = Graphics.FromImage(picImage.Image);

                while (_captureFlg)
                {
                    capture.Read(mat);

                    if (_saveFlg == false)
                    {
                        if (_bmpCamera != null) _bmpCamera.Dispose();
                        _bmpCamera = OpenCvSharp.Extensions.BitmapConverter.ToBitmap(mat);

                        // ROI領域の描画
                        _grpDisp.DrawImage(_bmpCamera, 0, 0, _bmpCamera.Width, _bmpCamera.Height);
                        _grpDisp.DrawRectangle(Pens.Red, _rect);

                        picImage.Refresh();
                    }

                    System.Threading.Thread.Sleep(30);
                    System.Windows.Forms.Application.DoEvents();
                }
            }
        }


        ///////////////////////////////////////////////////////////////
        //
        // メニューイベント
        //
        ///////////////////////////////////////////////////////////////

        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        ///////////////////////////////////////////////////////////////
        //
        // ボタンイベント
        //
        ///////////////////////////////////////////////////////////////

        /// <summary>
        /// 保存先フォルダの指定
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnSaveDir_Click(object sender, EventArgs e)
        {
            var fbd = new FolderBrowserDialog();
            fbd.RootFolder = Environment.SpecialFolder.Desktop;
            fbd.Description = "Select File Save Folder";
            if (fbd.ShowDialog(this) == DialogResult.OK)
            {
                txtSaveFolder.Text = fbd.SelectedPath;
            }
        }

        /// <summary>
        /// リストのクリア
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnClearList_Click(object sender, EventArgs e)
        {
            _imageIndex = 0;

            listView.Clear();
            imageList.Images.Clear();
        }

        ///////////////////////////////////////////////////////////////
        //
        // イベント
        //
        ///////////////////////////////////////////////////////////////

        private void picImage_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                // マウスの右ボタンの時は画像の撮影

                _saveFlg = true;

                // カメラ画像から指定されたROI領域のBitmapを切り出し
                var dest = _bmpCamera.Clone(_rect, _bmpCamera.PixelFormat);

                // ListViewへ画像の追加、ファイル保存
                AddSaveImage(dest);

                ///////////////////////////////////////////////////
                // 画像の水増し
                ///////////////////////////////////////////////////

                // 90°回転
                if (chbRot90.Checked == true)
                {
                    BitmapGenerate(dest, RotateFlipType.Rotate90FlipNone);
                }
                // 180°回転
                if (chbRot180.Checked == true)
                {
                    BitmapGenerate(dest, RotateFlipType.Rotate180FlipNone);
                }
                // 270°回転
                if (chbRot270.Checked == true)
                {
                    BitmapGenerate(dest, RotateFlipType.Rotate270FlipNone);
                }

                // 左右反転
                if (chbFlipX.Checked == true)
                {
                    BitmapGenerate(dest, RotateFlipType.RotateNoneFlipX);
                }
                // 上下反転
                if (chbFlipY.Checked == true)
                {
                    BitmapGenerate(dest, RotateFlipType.RotateNoneFlipY);
                }
                _saveFlg = false;
            }
            else
            {
                // マウスの左ボタンの時は撮影領域の移動
                SetROI(e.X, e.Y);
            }
        }

        ///////////////////////////////////////////////////////////////
        //
        // メソッド
        //
        ///////////////////////////////////////////////////////////////

        /// <summary>
        /// 画像の回転、反転を行い、リストへ追加、画像のファイル保存を行う
        /// </summary>
        /// <param name="original"></param>
        /// <param name="type"></param>
        private void BitmapGenerate(Bitmap original, RotateFlipType type)
        {
            // オリジナル画像のコピー
            var bmpGenerate = (Bitmap)original.Clone();
            // 画像の回転、反転
            bmpGenerate.RotateFlip(type);
            // リストへ追加、保存
            AddSaveImage(bmpGenerate);
            // 解放
            bmpGenerate.Dispose();
        }


        /// <summary>
        /// ListViewへ画像の追加、ファイル保存
        /// </summary>
        /// <param name="bmp"></param>
        private void AddSaveImage(Bitmap bmp)
        {
            // 保存先
            var filePath = txtSaveFolder.Text + "\\";
            // 保存するファイル名（Index番号より前の部分）
            var fileName = txtFilePrefix.Text + _imageIndex.ToString("D5");

            // フォルダが存在するか？
            if (System.IO.Directory.Exists(filePath) == false)
            {
                // フォルダが存在しない場合は作成する
                System.IO.Directory.CreateDirectory(filePath);
            }

            imageList.Images.Add(bmp);
            listView.Items.Add(fileName, imageList.Images.Count - 1);
            // 最後の行を表示
            listView.EnsureVisible(imageList.Images.Count - 1);

            SaveImage(bmp, filePath + fileName, cmbFileType.Text);

            // 次のIndexへ
            _imageIndex++;

        }

        /// <summary>
        /// 撮影領域の設定
        /// </summary>
        /// <param name="centerX">領域の中心のX座標</param>
        /// <param name="centerY">領域の中心のY座標</param>
        private void SetROI(int centerX, int centerY)
        {
            int roiWidth = (int)nudWidth.Value;
            int roiHeight = (int)nudHeight.Value;

            if (roiWidth > picImage.Width) roiWidth = picImage.Width;
            if (roiHeight > picImage.Height) roiHeight = picImage.Height;

            _rect = new Rectangle(centerX - roiWidth / 2, centerY - roiHeight / 2, roiWidth, roiHeight);

            if (_rect.Left < 0) _rect.X = 0;
            if (_rect.Top < 0) _rect.Y = 0;

            if (_rect.Right > picImage.Width - 1) _rect.X = picImage.Width - roiWidth;
            if (_rect.Bottom > picImage.Height - 1) _rect.Y = picImage.Height - roiHeight;
        }

        /// <summary>
        /// Bitmapオブジェクトのファイル保存
        /// </summary>
        /// <param name="bmp">Bitmapオブジェクト</param>
        /// <param name="fileName">保存するファイル名（拡張子なし）</param>
        /// <param name="fileType">ファイル形式の文字列（jpg,bmp, png, tifのいづれか）</param>
        private void SaveImage(Bitmap bmp, string fileName, string fileType)
        {
            switch (fileType)
            {
                case "jpg":
                    bmp.Save(fileName + ".jpg", System.Drawing.Imaging.ImageFormat.Jpeg);
                    break;

                case "bmp":
                    bmp.Save(fileName + ".bmp", System.Drawing.Imaging.ImageFormat.Bmp);
                    break;

                case "png":
                    bmp.Save(fileName + ".png", System.Drawing.Imaging.ImageFormat.Png);
                    break;

                case "tif":
                    bmp.Save(fileName + ".tif", System.Drawing.Imaging.ImageFormat.Tiff);
                    break;

                default:
                    break;
            }
        }
    }
}
